@@ -40,6 +40,10 @@ struct xatmi_exception : public std::exception {
   int code_;
   std::string message_;
 
+ protected:
+  xatmi_exception(int code, const std::string &message)
+      : code_(code), message_(message) {}
+
  public:
   explicit xatmi_exception(int code)
       : code_(code), message_(tpstrerror(code)) {}
@@ -48,6 +52,49 @@ struct xatmi_exception : public std::exception {
   int code() const noexcept { return code_; }
 };
 
+struct qm_exception : public xatmi_exception {
+ public:
+  explicit qm_exception(int code) : xatmi_exception(code, qmstrerror(code)) {}
+
+  static const char *qmstrerror(int code) {
+    switch (code) {
+      case QMEINVAL:
+        return "An invalid flag value was specified.";
+      case QMEBADRMID:
+        return "An invalid resource manager identifier was specified.";
+      case QMENOTOPEN:
+        return "The resource manager is not currently open.";
+      case QMETRAN:
+        return "Transaction error.";
+      case QMEBADMSGID:
+        return "An invalid message identifier was specified.";
+      case QMESYSTEM:
+        return "A system error occurred. The exact nature of the error is "
+               "written to a log file.";
+      case QMEOS:
+        return "An operating system error occurred.";
+      case QMEABORTED:
+        return "The operation was aborted.";
+      case QMEPROTO:
+        return "An enqueue was done when the transaction state was not active.";
+      case QMEBADQUEUE:
+        return "An invalid or deleted queue name was specified.";
+      case QMENOSPACE:
+        return "Insufficient resources.";
+      case QMERELEASE:
+        return "Unsupported feature.";
+      case QMESHARE:
+        return "Queue is opened exclusively by another application.";
+      case QMENOMSG:
+        return "No message was available for dequeuing.";
+      case QMEINUSE:
+        return "Message is in use by another transaction.";
+      default:
+        return "?";
+    }
+  }
+};
+
 struct fml32_exception : public std::exception {
  private:
   int code_;
@@ -61,68 +108,23 @@ struct fml32_exception : public std::exception {
   int code() const noexcept { return code_; }
 };
 
-static PyObject *TuxedoException_tp_str(PyObject *selfPtr) {
-  py::str ret;
-  try {
-    py::handle self(selfPtr);
-    py::tuple args = self.attr("args");
-    ret = py::str(args[0]);
-  } catch (py::error_already_set &e) {
-    ret = "";
-  }
-
-  ret.inc_ref();
-  return ret.ptr();
-}
-
-static PyObject *TuxedoException_code(PyObject *selfPtr, void *closure) {
-  try {
-    py::handle self(selfPtr);
-    py::tuple args = self.attr("args");
-    py::object code = args[1];
-    code.inc_ref();
-    return code.ptr();
-  } catch (py::error_already_set &e) {
-    py::none ret;
-    ret.inc_ref();
-    return ret.ptr();
-  }
-}
-
-static PyGetSetDef TuxedoException_getsetters[] = {
-    {const_cast<char *>("code"), TuxedoException_code, nullptr, nullptr,
-     nullptr},
-    {nullptr}};
-
-struct client {
-  client() {}
-  // TODO: real params
-  explicit client(bool client) {
-    TPCONTEXT_T context;
-    if (tpgetctxt(&context, 0) != -1 && context >= 0) {
+struct context {
+  context() {}
+  explicit context(bool is_client) {
+    TPCONTEXT_T tpcontext;
+    if (tpgetctxt(&tpcontext, 0) != -1 && tpcontext >= 0) {
       return;
     }
-    _init(
-        [client](TPINIT *tpinfo) {
-          if (client) {
-            if (tpinit(tpinfo) == -1) {
-              throw xatmi_exception(tperrno);
-            }
-          } else {
-            if (tpappthrinit(tpinfo) == -1) {
-              throw xatmi_exception(tperrno);
-            }
-          }
-        },
-        nullptr, "tpsysadm", nullptr, nullptr, TPMULTICONTEXTS);
+    _init(is_client ? tpinit : tpappthrinit, nullptr, "tpsysadm", nullptr,
+          nullptr, TPMULTICONTEXTS);
   }
-  client(std::function<void(TPINIT *)> f, const char *usrname,
-         const char *cltname, const char *passwd, const char *grpname,
-         long flags) {
-    _init(f, usrname, cltname, passwd, grpname, flags);
+  context(std::function<int(TPINIT *)> initfunc, const char *usrname,
+          const char *cltname, const char *passwd, const char *grpname,
+          long flags) {
+    _init(initfunc, usrname, cltname, passwd, grpname, flags);
   }
 
-  void _init(std::function<void(TPINIT *)> f, const char *usrname,
+  void _init(std::function<int(TPINIT *)> initfunc, const char *usrname,
              const char *cltname, const char *passwd, const char *grpname,
              long flags) {
     std::unique_ptr<char, decltype(&tpfree)> guard(
@@ -144,14 +146,16 @@ struct client {
       strncpy(tpinfo->grpname, grpname, sizeof(tpinfo->grpname));
     }
     tpinfo->flags = flags;
-    f(tpinfo);
+    if (initfunc(tpinfo) == -1) {
+      throw xatmi_exception(tperrno);
+    }
   }
 
-  ~client() {}
-  client(const client &) = delete;
-  client &operator=(const client &) = delete;
-  client(client &&) = delete;
-  client &operator=(client &&) = delete;
+  ~context() {}
+  context(const context &) = delete;
+  context &operator=(const context &) = delete;
+  context(context &&) = delete;
+  context &operator=(context &&) = delete;
 };
 
 struct svcresult {
@@ -244,12 +248,12 @@ struct pytpreply {
 };
 
 static py::object server;
-static thread_local std::unique_ptr<client> tclient;
+static thread_local std::unique_ptr<context> thread_context;
 static thread_local svcresult tsvcresult;
 
-static void default_client() {
-  if (!tclient) {
-    tclient.reset(new client(server.ptr() == nullptr));
+static void with_context() {
+  if (!thread_context) {
+    thread_context.reset(new context(server.ptr() == nullptr));
   }
 }
 
@@ -509,7 +513,7 @@ static void pytppost(const std::string eventname, py::object data, long flags) {
 }
 
 static pytpreply pytpcall(const char *svc, py::object idata, long flags) {
-  default_client();
+  with_context();
   auto in = from_py(idata);
   xatmibuf out("FML32", 1024);
   {
@@ -527,13 +531,16 @@ static pytpreply pytpcall(const char *svc, py::object idata, long flags) {
 
 static TPQCTL pytpenqueue(const char *qspace, const char *qname, TPQCTL *ctl,
                           py::object data, long flags) {
-  default_client();
+  with_context();
   auto in = from_py(data);
   {
     py::gil_scoped_release release;
     int rc = tpenqueue(const_cast<char *>(qspace), const_cast<char *>(qname),
                        ctl, *in.pp, in.len, flags);
     if (rc == -1) {
+      if (tperrno == TPEDIAGNOSTIC) {
+        throw qm_exception(ctl->diagnostic);
+      }
       throw xatmi_exception(tperrno);
     }
   }
@@ -543,13 +550,16 @@ static TPQCTL pytpenqueue(const char *qspace, const char *qname, TPQCTL *ctl,
 static std::pair<TPQCTL, py::object> pytpdequeue(const char *qspace,
                                                  const char *qname, TPQCTL *ctl,
                                                  long flags) {
-  default_client();
+  with_context();
   xatmibuf out("FML32", 1024);
   {
     py::gil_scoped_release release;
     int rc = tpdequeue(const_cast<char *>(qspace), const_cast<char *>(qname),
                        ctl, out.pp, &out.len, flags);
     if (rc == -1) {
+      if (tperrno == TPEDIAGNOSTIC) {
+        throw qm_exception(ctl->diagnostic);
+      }
       throw xatmi_exception(tperrno);
     }
   }
@@ -572,7 +582,7 @@ static pytpreply pytpadmcall(py::object idata, long flags) {
 }
 
 static int pytpacall(const char *svc, py::object idata, long flags) {
-  default_client();
+  with_context();
   auto in = from_py(idata);
 
   py::gil_scoped_release release;
@@ -584,7 +594,7 @@ static int pytpacall(const char *svc, py::object idata, long flags) {
 }
 
 static pytpreply pytpgetrply(int cd, long flags) {
-  default_client();
+  with_context();
   xatmibuf out("FML32", 1024);
   {
     py::gil_scoped_release release;
@@ -599,8 +609,8 @@ static pytpreply pytpgetrply(int cd, long flags) {
 }
 
 int tpsvrinit(int argc, char *argv[]) {
-  if (!tclient) {
-    tclient.reset(new client());
+  if (!thread_context) {
+    thread_context.reset(new context());
   }
   if (tpopen() == -1) {
     userlog(const_cast<char *>("Failed tpopen() = %d / %s"), tperrno,
@@ -624,9 +634,15 @@ void tpsvrdone() {
   }
 }
 int tpsvrthrinit(int argc, char *argv[]) {
-  if (!tclient) {
-    tclient.reset(new client());
+  if (!thread_context) {
+    thread_context.reset(new context());
   }
+  // Create a new Python thread
+  // otherwise pybind11 creates and deletes one
+  // and messes up threading.local
+  auto const &internals = pybind11::detail::get_internals();
+  PyThreadState_New(internals.istate);
+
   if (tpopen() == -1) {
     userlog(const_cast<char *>("Failed tpopen() = %d / %s"), tperrno,
             tpstrerror(tperrno));
@@ -649,8 +665,8 @@ void tpsvrthrdone() {
   }
 }
 void PY(TPSVCINFO *svcinfo) {
-  if (!tclient) {
-    tclient.reset(new client());
+  if (!thread_context) {
+    thread_context.reset(new context());
   }
   tsvcresult.clean = true;
 
@@ -713,7 +729,6 @@ static void pytpadvertisex(std::string svcname, long flags) {
 
 extern "C" {
 int _tmrunserver(int);
-extern struct xa_switch_t xaosw;
 extern struct xa_switch_t tmnull_switch;
 extern int _tmbuilt_with_thread_option;
 }
@@ -731,7 +746,6 @@ static xao_svc_ctx *xao_svc_ctx_ptr;
 struct tmsvrargs_t *_tmgetsvrargs(const char *rmname) {
   tmsvrargs.reserved1 = nullptr;
   tmsvrargs.reserved2 = nullptr;
-  // tmsvrargs.xa_switch = &xaosw;
   if (strcasecmp(rmname, "NONE") == 0) {
     tmsvrargs.xa_switch = &tmnull_switch;
 #if defined(_WIN32) || defined(_WIN64)
@@ -781,7 +795,106 @@ static void pyrun(py::object svr, std::vector<std::string> args,
   }
 }
 
+static PyObject *TuxedoException_code(PyObject *selfPtr, void *closure) {
+  try {
+    py::handle self(selfPtr);
+    py::tuple args = self.attr("args");
+    py::object code = args[1];
+    code.inc_ref();
+    return code.ptr();
+  } catch (py::error_already_set &e) {
+    py::none ret;
+    ret.inc_ref();
+    return ret.ptr();
+  }
+}
+
+static PyGetSetDef TuxedoException_getsetters[] = {
+    {const_cast<char *>("code"), TuxedoException_code, nullptr, nullptr,
+     nullptr},
+    {nullptr}};
+
+static PyObject *TuxedoException_tp_str(PyObject *selfPtr) {
+  py::str ret;
+  try {
+    py::handle self(selfPtr);
+    py::tuple args = self.attr("args");
+    ret = py::str(args[0]);
+  } catch (py::error_already_set &e) {
+    ret = "";
+  }
+
+  ret.inc_ref();
+  return ret.ptr();
+}
+
+static void register_exceptions(py::module &m) {
+  static PyObject *XatmiException =
+      PyErr_NewException("tuxedo.XatmiException", nullptr, nullptr);
+  if (XatmiException) {
+    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(XatmiException);
+    as_type->tp_str = TuxedoException_tp_str;
+    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
+    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
+    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
+
+    Py_XINCREF(XatmiException);
+    m.add_object("XatmiException", py::handle(XatmiException));
+  }
+
+  static PyObject *QmException =
+      PyErr_NewException("tuxedo.QmException", nullptr, nullptr);
+  if (QmException) {
+    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(QmException);
+    as_type->tp_str = TuxedoException_tp_str;
+    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
+    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
+    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
+
+    Py_XINCREF(QmException);
+    m.add_object("QmException", py::handle(QmException));
+  }
+
+  static PyObject *Fml32Exception =
+      PyErr_NewException("tuxedo.Fml32Exception", nullptr, nullptr);
+  if (Fml32Exception) {
+    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(Fml32Exception);
+    as_type->tp_str = TuxedoException_tp_str;
+    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
+    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
+    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
+
+    Py_XINCREF(Fml32Exception);
+    m.add_object("Fml32Exception", py::handle(Fml32Exception));
+  }
+
+  py::register_exception_translator([](std::exception_ptr p) {
+    try {
+      if (p) {
+        std::rethrow_exception(p);
+      }
+    } catch (const qm_exception &e) {
+      py::tuple args(2);
+      args[0] = e.what();
+      args[1] = e.code();
+      PyErr_SetObject(QmException, args.ptr());
+    } catch (const xatmi_exception &e) {
+      py::tuple args(2);
+      args[0] = e.what();
+      args[1] = e.code();
+      PyErr_SetObject(XatmiException, args.ptr());
+    } catch (const fml32_exception &e) {
+      py::tuple args(2);
+      args[0] = e.what();
+      args[1] = e.code();
+      PyErr_SetObject(Fml32Exception, args.ptr());
+    }
+  });
+}
+
 PYBIND11_MODULE(tuxedo, m) {
+  register_exceptions(m);
+
   // Poor man's namedtuple
   py::class_<pytpreply>(m, "TpReply")
       .def_readonly("rval", &pytpreply::rval)
@@ -850,50 +963,6 @@ PYBIND11_MODULE(tuxedo, m) {
       .def_readonly("delivery_qos", &TPQCTL::delivery_qos)
       .def_readonly("reply_qos", &TPQCTL::reply_qos);
 
-  static PyObject *XatmiException =
-      PyErr_NewException("tuxedo.XatmiException", nullptr, nullptr);
-
-  if (XatmiException) {
-    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(XatmiException);
-    as_type->tp_str = TuxedoException_tp_str;
-    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
-    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
-    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
-
-    Py_XINCREF(XatmiException);
-    m.add_object("XatmiException", py::handle(XatmiException));
-  }
-  static PyObject *Fml32Exception =
-      PyErr_NewException("tuxedo.Fml32Exception", nullptr, nullptr);
-  if (Fml32Exception) {
-    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(Fml32Exception);
-    as_type->tp_str = TuxedoException_tp_str;
-    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
-    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
-    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
-
-    Py_XINCREF(Fml32Exception);
-    m.add_object("Fml32Exception", py::handle(Fml32Exception));
-  }
-
-  py::register_exception_translator([](std::exception_ptr p) {
-    try {
-      if (p) {
-        std::rethrow_exception(p);
-      }
-    } catch (const xatmi_exception &e) {
-      py::tuple args(2);
-      args[0] = e.what();
-      args[1] = e.code();
-      PyErr_SetObject(XatmiException, args.ptr());
-    } catch (const fml32_exception &e) {
-      py::tuple args(2);
-      args[0] = e.what();
-      args[1] = e.code();
-      PyErr_SetObject(Fml32Exception, args.ptr());
-    }
-  });
-
   m.def(
       "xaoSvcCtx",
       []() {
@@ -910,13 +979,8 @@ PYBIND11_MODULE(tuxedo, m) {
       [](const char *usrname, const char *cltname, const char *passwd,
          const char *grpname, long flags) {
         py::gil_scoped_release release;
-        tclient.reset(new client(
-            [](TPINIT *tpinfo) {
-              if (tpinit(tpinfo) == -1) {
-                throw xatmi_exception(tperrno);
-              }
-            },
-            usrname, cltname, passwd, grpname, flags));
+        thread_context.reset(
+            new context(tpinit, usrname, cltname, passwd, grpname, flags));
       },
       "Joins an application", py::arg("usrname") = nullptr,
       py::arg("cltname") = nullptr, py::arg("passwd") = nullptr,
@@ -926,7 +990,7 @@ PYBIND11_MODULE(tuxedo, m) {
       "tpterm",
       []() {
         py::gil_scoped_release release;
-        tclient.reset();
+        thread_context.reset();
         if (tpterm() == -1) {
           throw xatmi_exception(tperrno);
         }
@@ -938,29 +1002,25 @@ PYBIND11_MODULE(tuxedo, m) {
       [](const char *usrname, const char *cltname, const char *passwd,
          const char *grpname, long flags) {
         py::gil_scoped_release release;
-        tclient.reset(new client(
-            [](TPINIT *tpinfo) {
-              if (tpappthrinit(tpinfo) == -1) {
-                throw xatmi_exception(tperrno);
-              }
-            },
-            usrname, cltname, passwd, grpname, flags));
+        thread_context.reset(new context(tpappthrinit, usrname, cltname, passwd,
+                                         grpname, flags));
       },
-      "Joins an application", py::arg("usrname") = nullptr,
-      py::arg("cltname") = nullptr, py::arg("passwd") = nullptr,
-      py::arg("grpname") = nullptr, py::arg("flags") = 0);
+      "Routine for creating and initializing a new Tuxedo context in an "
+      "application-created server thread.",
+      py::arg("usrname") = nullptr, py::arg("cltname") = nullptr,
+      py::arg("passwd") = nullptr, py::arg("grpname") = nullptr,
+      py::arg("flags") = 0);
 
   m.def(
       "tpappthrterm",
       []() {
         py::gil_scoped_release release;
-        tclient.reset();
+        thread_context.reset();
         if (tpappthrterm() == -1) {
           throw xatmi_exception(tperrno);
         }
       },
-      "Routine for creating and initializing a new Tuxedo context in an "
-      "application-created server thread.");
+      "Routine for terminating Tuxedo User context in a server process");
 
   m.def(
       "tpbegin",
@@ -1099,6 +1159,7 @@ PYBIND11_MODULE(tuxedo, m) {
   m.def(
       "tpgblktime",
       [](long flags) {
+        with_context();
         int rc = tpgblktime(flags);
         if (rc == -1) {
           throw xatmi_exception(tperrno);
@@ -1111,6 +1172,7 @@ PYBIND11_MODULE(tuxedo, m) {
   m.def(
       "tpsblktime",
       [](int blktime, long flags) {
+        with_context();
         if (tpsblktime(blktime, flags) == -1) {
           throw xatmi_exception(tperrno);
         }
@@ -1275,6 +1337,24 @@ PYBIND11_MODULE(tuxedo, m) {
   m.attr("TPENOSECONDARYRQ") = py::int_(TPENOSECONDARYRQ);
 #endif
 
+  m.attr("QMEINVAL") = py::int_(QMEINVAL);
+  m.attr("QMEBADRMID") = py::int_(QMEBADRMID);
+  m.attr("QMENOTOPEN") = py::int_(QMENOTOPEN);
+  m.attr("QMETRAN") = py::int_(QMETRAN);
+  m.attr("QMEBADMSGID") = py::int_(QMEBADMSGID);
+  m.attr("QMESYSTEM") = py::int_(QMESYSTEM);
+  m.attr("QMEOS") = py::int_(QMEOS);
+  m.attr("QMEABORTED") = py::int_(QMEABORTED);
+  m.attr("QMENOTA") = py::int_(QMENOTA);
+  m.attr("QMEPROTO") = py::int_(QMEPROTO);
+  m.attr("QMEBADQUEUE") = py::int_(QMEBADQUEUE);
+  m.attr("QMENOMSG") = py::int_(QMENOMSG);
+  m.attr("QMEINUSE") = py::int_(QMEINUSE);
+  m.attr("QMENOSPACE") = py::int_(QMENOSPACE);
+  m.attr("QMERELEASE") = py::int_(QMERELEASE);
+  m.attr("QMEINVHANDLE") = py::int_(QMEINVHANDLE);
+  m.attr("QMESHARE") = py::int_(QMESHARE);
+
   m.attr("FLD_SHORT") = py::int_(FLD_SHORT);
   m.attr("FLD_LONG") = py::int_(FLD_LONG);
   m.attr("FLD_CHAR") = py::int_(FLD_CHAR);
